/* Viterbi decoder for K=7 rate=1/2 convolutional code - chainback version
 * Copyright 1996 Phil Karn, KA9Q
 */
#include "rcpc.h"

extern unsigned char Partab[];	/* Parity lookup table */


/* Puts the received data in the right places so that a usual VA
 * can be used (unused places should be initialized as erasures)
 * This effectively 'unpunctures' the received code stream 
 * This program has to be called successively for l=0,1,2...
 * BEFORE using 'viterbi' decoder
 */
pack_symbols(
unsigned char *new_symbols, /* New symbols obtained at step l */
unsigned char *symbols,	    /* array with depunctured symbols */
unsigned int l,             /* l is at what step we are*/
unsigned int nbytes         /* number of UNCODED DATA bytes to process */
){

  int i, packed=0;
  unsigned char encstate, a0, a1, a2; /*1 indicate bits received at step l*/

  /* we first create new puncturing matrix which has ones only in
     positions which have to be sent at iteration 'l'. For l=0 this
     corresponds   to the highest rate code puncturing matrix. At each 
     consequitive step only additional bits are sent */
 
    if(l==0) {/*first transmission. initialize 'symbols'*/
      a0=A0[0];
      a1=A1[0];
      a2=A2[0];
      memset(symbols,OFFSET,nbytes*RATE*8);
    }
    
    else { /* l > 0 */
      a0=A0[l]^A0[l-1];
      a1=A1[l]^A1[l-1];
      a2=A2[l]^A2[l-1];
    }
    
    while( nbytes-- != 0){
      for(i=7 ; i>=0 ; i--) {
	if((a0 >> i) & 1){
	  *symbols = *new_symbols++;
	  packed ++;
	}
	symbols++;

	if((a1 >> i) & 1){
	  *symbols = *new_symbols++;
	  packed ++;
	}
	symbols++;

	if((a2 >> i) & 1){
	  *symbols = *new_symbols++;
	  packed++;
	}
	symbols++;
      }
    }   
    return packed;
}





/* Viterbi decoder */
int
viterbi(
unsigned long *metric,	/* Final path metric (returned value) */
unsigned char *data,	/* Decoded output data */
unsigned char *symbols,	/* Raw deinterleaved input symbols */
unsigned int nbits,	/* Number of output bits */
int mettab[2][256],	/* Metric table, [sent sym][rx symbol] */
int PATH                /* return path number in list VA */  
){
	unsigned int bitcnt = 0;
	int beststate,state,i,l;
	static long cmetric[16*L],nmetric[16*L];
	unsigned char *paths;
	unsigned char *pp, *ppp, pp0;
	unsigned char dec[L];
	int mets[8];
	int npaths; /*number of distinct paths at each node - to warm up*/
	unsigned int path_offset;


	paths = (unsigned char *)malloc(nbits*16*L);
	pp = paths;
	path_offset=nbits*16; /*offset between paths */

	npaths=1;
	/* Initialize starting metrics to prefer 0 state */
	for(i=0;i<16;i++) {
		cmetric[i] = -1000000;
	}
	cmetric[0] = 0;

	for(;;){

		/* Read input symbols and compute branch metrics */
     mets[0]=mettab[0][symbols[0]]+mettab[0][symbols[1]]+mettab[0][symbols[2]];
     mets[1]=mettab[0][symbols[0]]+mettab[0][symbols[1]]+mettab[1][symbols[2]];
     mets[2]=mettab[0][symbols[0]]+mettab[1][symbols[1]]+mettab[0][symbols[2]];
     mets[3]=mettab[0][symbols[0]]+mettab[1][symbols[1]]+mettab[1][symbols[2]];
     mets[4]=mettab[1][symbols[0]]+mettab[0][symbols[1]]+mettab[0][symbols[2]];
     mets[5]=mettab[1][symbols[0]]+mettab[0][symbols[1]]+mettab[1][symbols[2]];
     mets[6]=mettab[1][symbols[0]]+mettab[1][symbols[1]]+mettab[0][symbols[2]];
     mets[7]=mettab[1][symbols[0]]+mettab[1][symbols[1]]+mettab[1][symbols[2]];
		symbols += 3;

	  BUTTERFLY(0,0,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+0) = dec[l];

	  BUTTERFLYx(0,0,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+1) = dec[l];

	  BUTTERFLY(1,5,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+2) = dec[l];

	  BUTTERFLYx(1,5,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+3) = dec[l];

	  BUTTERFLY(2,3,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+4) = dec[l];

	  BUTTERFLYx(2,3,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+5) = dec[l];

	  BUTTERFLY(3,6,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+6) = dec[l];

	  BUTTERFLYx(3,6,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+7) = dec[l];

	  BUTTERFLY(4,2,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+8) = dec[l];

	  BUTTERFLYx(4,2,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+9) = dec[l];

	  BUTTERFLY(5,7,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+10) = dec[l];

	  BUTTERFLYx(5,7,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+11) = dec[l];

	  BUTTERFLY(6,1,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+12) = dec[l];

	  BUTTERFLYx(6,1,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+13) = dec[l];

	  BUTTERFLY(7,4,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+14) = dec[l];

	  BUTTERFLYx(7,4,cmetric,nmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+15) = dec[l];


                pp += 16;



		if(++bitcnt == nbits){
		        beststate = 0;
			for(l=0;l<L;l++)
			  metric[l] = nmetric[beststate+l*16];
			break;
		}
		npaths=(2*npaths>L)? L:2*npaths;

		/* Read input symbols and compute branch metrics */
     mets[0]=mettab[0][symbols[0]]+mettab[0][symbols[1]]+mettab[0][symbols[2]];
     mets[1]=mettab[0][symbols[0]]+mettab[0][symbols[1]]+mettab[1][symbols[2]];
     mets[2]=mettab[0][symbols[0]]+mettab[1][symbols[1]]+mettab[0][symbols[2]];
     mets[3]=mettab[0][symbols[0]]+mettab[1][symbols[1]]+mettab[1][symbols[2]];
     mets[4]=mettab[1][symbols[0]]+mettab[0][symbols[1]]+mettab[0][symbols[2]];
     mets[5]=mettab[1][symbols[0]]+mettab[0][symbols[1]]+mettab[1][symbols[2]];
     mets[6]=mettab[1][symbols[0]]+mettab[1][symbols[1]]+mettab[0][symbols[2]];
     mets[7]=mettab[1][symbols[0]]+mettab[1][symbols[1]]+mettab[1][symbols[2]];
		symbols += 3;


	  BUTTERFLY(0,0,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+0) = dec[l];

	  BUTTERFLYx(0,0,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+1) = dec[l];

	  BUTTERFLY(1,5,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+2) = dec[l];

	  BUTTERFLYx(1,5,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+3) = dec[l];

	  BUTTERFLY(2,3,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+4) = dec[l];

	  BUTTERFLYx(2,3,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+5) = dec[l];

	  BUTTERFLY(3,6,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+6) = dec[l];

	  BUTTERFLYx(3,6,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+7) = dec[l];

	  BUTTERFLY(4,2,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+8) = dec[l];

	  BUTTERFLYx(4,2,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+9) = dec[l];

	  BUTTERFLY(5,7,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+10) = dec[l];

	  BUTTERFLYx(5,7,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+11) = dec[l];

	  BUTTERFLY(6,1,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+12) = dec[l];

	  BUTTERFLYx(6,1,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+13) = dec[l];

	  BUTTERFLY(7,4,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+14) = dec[l];

	  BUTTERFLYx(7,4,nmetric,cmetric,mets,dec,npaths);
                for(l=0;l<L;l++)
                  *(pp+l*path_offset+15) = dec[l];


                pp += 16;



		if(++bitcnt == nbits){
		        beststate = 0;
			for(l=0;l<L;l++)
			  metric[l] = cmetric[beststate+l*16];
			break;
		}
		npaths=(2*npaths>L)? L:2*npaths;


		/*
		beststate=0;
		for(i=0 ; i<16 ;i++)
		  if(cmetric[beststate] < cmetric[i])
		    beststate = i;
		traceback(beststate,0,pp,path_offset,data,bitcnt);
		for(i=0;i<bitcnt/8;i++)
		  printf("%02x",data[i]);
		printf("\n");
		*/

	}

	pp -= 16; /* undo the last path increment */
	pp += PATH*path_offset; /* chose the correct path */
	/* Chain back from terminal state to produce decoded data
	 * i>>3 is to pack bits in one byte symbol
	 * This part doesn't change if we take another code rate 
	 */
	memset(data,0,nbits/8);
	for(i=nbits-5;i >= 0;i--){
		state = beststate;
		pp0 = pp[state];
		if(pp0 >= 128){ /*we came from the lower branch*/
			beststate |= 16;
			data[i>>3] |= 0x80 >> (i&7);
		}

		beststate >>= 1;
		pp+=path_offset*((pp0 & 0x7f)-PATH)-16;
		PATH = pp0 & 0x7f;

	}
	free(paths);
	return 0;
}

/* two butterflies 
 * first for even terminal state
 * second--for odd
 */

int 
BUTTERFLY(
int i,
int sym,
long *cmetric,
long *nmetric,
int  *mets,
unsigned char *dec,
int  npaths) { 
        int m0,m1,s,new_npaths, s0, s1,d,d0,d1;
	
	new_npaths = (2*npaths>L)? L:2*npaths;
	s0=0;
	s1=0;

	for(s=0; s<new_npaths ; s++) {
	    d0=16*s0+i; d1=16*s1+i+8; d=16*s+2*i;
	    if(s0<npaths) /* s0 may go out of range */
	      m0 = cmetric[d0] + mets[sym];
	    else
	      m0 = -1000000;
	    if(s1<npaths)
	      m1 = cmetric[d1] + mets[7^sym];
	    else
	      m1 = -1000000;


	    if((s0>=npaths) || ((m1 > m0) && s1<npaths)){
	 	  nmetric[d] = m1;
		  dec[s] = 128+s1;
		  s1++;
	    }
	    else {
	          nmetric[d] = m0;
		  dec[s] = s0;
		  s0++;
	    }
	}
}

int 
BUTTERFLYx(
int i,
int sym,
long *cmetric,
long *nmetric,
int  *mets,
unsigned char *dec,
int  npaths) { 
        int m0,m1,s,new_npaths, s0, s1,d,d0,d1;
	
	new_npaths = (2*npaths>L)? L:2*npaths;
	s0=0;
	s1=0;

	for(s=0; s<new_npaths ; s++) {
	    d0=16*s0+i; d1=16*s1+i+8; d=16*s+2*i+1;
	    if(s0<npaths)
	      m0 = cmetric[d0] + mets[7^sym];
	    else
	      m0 = -1000000;
	    if(s1<npaths)
	      m1 = cmetric[d1] + mets[sym] ;
	    else
	      m1 = -1000000;

	    if((s0>=npaths) || ((m1 > m0) && s1<npaths)){
		  nmetric[d] = m1;
		  dec[s] = 128 + s1;
		  s1++;
	    }
	    else {
	          nmetric[d] = m0;
		  dec[s] = s0;
		  s0++;
	    }
	}
}






/* used to traceback paths */
int
traceback(
int beststate,
int path, 
unsigned char *pp,
int path_offset,
unsigned char *data,
int nbits
) {

  int i,state;
  unsigned char pp0;
  
	pp -= 16; /* undo the last path increment */
	pp += path*path_offset; /* chose the correct path */
	/* Chain back from terminal state to produce decoded data
	 * i>>3 is to pack bits in one byte symbol
	 * This part doesn't change if we take another code rate 
	 */
	memset(data,0,nbits/8);
	for(i=nbits-5;i >= 0;i--){
		state = beststate;
		pp0 = pp[state];
		if(pp0 >= 128){ /*we came from the lower branch*/
			beststate |= 16;
			data[i>>3] |= 0x80 >> (i&7);
		}

		beststate >>= 1;
		pp+=path_offset*((pp0 & 0x7f)-path)-16;
		path = pp0 & 0x7f;
	}
	return(0);
}
