/* Test a rate 1/3 soft decision viterbi decoder */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rcpc.h"

void modnoise(unsigned char *,unsigned int,double,double);

/* Lookup table giving count of 1 bits for integers 0-255 */
unsigned char Bitcnt[] = {
 0, 1, 1, 2, 1, 2, 2, 3,
 1, 2, 2, 3, 2, 3, 3, 4,
 1, 2, 2, 3, 2, 3, 3, 4,
 2, 3, 3, 4, 3, 4, 4, 5,
 1, 2, 2, 3, 2, 3, 3, 4,
 2, 3, 3, 4, 3, 4, 4, 5,
 2, 3, 3, 4, 3, 4, 4, 5,
 3, 4, 4, 5, 4, 5, 5, 6,
 1, 2, 2, 3, 2, 3, 3, 4,
 2, 3, 3, 4, 3, 4, 4, 5,
 2, 3, 3, 4, 3, 4, 4, 5,
 3, 4, 4, 5, 4, 5, 5, 6,
 2, 3, 3, 4, 3, 4, 4, 5,
 3, 4, 4, 5, 4, 5, 5, 6,
 3, 4, 4, 5, 4, 5, 5, 6,
 4, 5, 5, 6, 5, 6, 6, 7,
 1, 2, 2, 3, 2, 3, 3, 4,
 2, 3, 3, 4, 3, 4, 4, 5,
 2, 3, 3, 4, 3, 4, 4, 5,
 3, 4, 4, 5, 4, 5, 5, 6,
 2, 3, 3, 4, 3, 4, 4, 5,
 3, 4, 4, 5, 4, 5, 5, 6,
 3, 4, 4, 5, 4, 5, 5, 6,
 4, 5, 5, 6, 5, 6, 6, 7,
 2, 3, 3, 4, 3, 4, 4, 5,
 3, 4, 4, 5, 4, 5, 5, 6,
 3, 4, 4, 5, 4, 5, 5, 6,
 4, 5, 5, 6, 5, 6, 6, 7,
 3, 4, 4, 5, 4, 5, 5, 6,
 4, 5, 5, 6, 5, 6, 6, 7,
 4, 5, 5, 6, 5, 6, 6, 7,
 5, 6, 6, 7, 6, 7, 7, 8,
};

main(argc,argv)
int argc;
char *argv[];
{
	double ebn0,esn0,noise;
	int mettab[2][256];
	int amp,nbits,i,ii,s;
	unsigned char *symbols, *new_symbols ;
	unsigned char *symbols1, *symbols2 ;
	unsigned char *data,*decdata;
	unsigned long metric[L];
	long t,ntrials;
	long seed;
	extern char *optarg;
	unsigned long biterrs = 0;
	unsigned long framerrs = 0;
	unsigned long corrected_frames = 0;
	unsigned long corrected_bits = 0, err;
	int timetrial;
        int P_RATE; /* punctured rate 0 -> (8/9), 8 -> (8/24) */

	P_RATE = 8; /*8 - rate 1/3; 0 - rate 8/9 */
	amp = 100;
	ebn0 = 4.7;
	nbits = 2040/*8*150*/; /*1152*/
	ntrials = 1000;
	timetrial = 0;
	time(&seed);

	while((i = getopt(argc,argv,"a:e:n:N:r:s:t")) != EOF){
		switch(i){
		case 'a':
			amp = atoi(optarg);	/* Signal amplitude in units */
			break;
		case 'e':
			ebn0 = atof(optarg);	/* Eb/N0 in dB */
			break;
		case 'n':
			nbits = atoi(optarg);	/* Number of data bits */
			break;
		case 'N':
			ntrials = atoi(optarg);
			break;
		case 's':
			seed = atoi(optarg);
			break;
		case 't':	/* Generate noise once, for decoder timing */
			timetrial = 1;
			break;
		case 'r':	/* Rate of the code to use */
			P_RATE = atoi(optarg);
			break;
		case '?':
			usage();
			exit(1);
		}
	}
	srandom(seed);

	esn0 = ebn0 + 10*log10(8.0/RR[P_RATE]);	/* Es/N0 in dB */	

	/* Compute noise voltage. The 0.5 factor accounts for BPSK seeing
	 * only half the noise power, and the sqrt() converts power to
	 * voltage.
	 */
	noise = sqrt(0.5/pow(10.,esn0/10.));
	
	data = malloc(1 + nbits/8);
	decdata = malloc(1 + nbits/8);
	new_symbols = malloc(3+nbits*3);
	symbols = malloc(3+nbits*3);

	symbols1 = malloc(3+nbits*3);
	symbols2 = malloc(3+nbits*3);

	/* Generate metrics analytically, with gaussian pdf */	
	gen_met(mettab,amp,noise,0.,4);
	printf("metric table range %d to %d\n",mettab[0][0],mettab[1][0]);

	/* Generate data */
	memset(data,0,1+nbits/8);	/* All 0's, including tail */
	data[0] = 0x80;	
	for(i=1 ; i<nbits/8-1; i++)
	  data[i] = 0x55;


	
	i = 0;
	/*
	if(timetrial){
		rcpc_encode(new_symbols,data,S,1+nbits/8);
		modnoise(new_symbols,(1+nbits/8)*RR[S],amp,noise);
		for(t = 1;t <= ntrials;t++)
		        pack_symbols(new_symbols,symbols,S,1+nbits/8);
			viterbi(&metric,decdata,symbols,nbits,mettab,0);
		printf("bits decoded = %ld\n",nbits*ntrials);
		exit(0);
	}
	*/
	for(t = 1;t <= ntrials;t++){
		/* Modulate and add noise */
	  
	     for(s=0 ; s<= P_RATE ; s++) {
		   rcpc_encode(new_symbols,data,s,1+nbits/8);
		   modnoise(new_symbols,(nbits/8+1)*RRR[s],amp,noise);
		   pack_symbols(new_symbols,symbols,s,1+nbits/8);
	     }
	     free(new_symbols); 

	     viterbi(metric,decdata,symbols,nbits,mettab,0);
	     if(memcmp(data,decdata,1+nbits/8) != 0){
		      printf("Frame %d metric %d decoded data:\n",t,metric[0]);
			for(i=0;i<1+nbits/8;i++)
				printf("%02x",decdata[i]);
			printf("\n");
			framerrs++;
			err=0;
			for(i=0;i<1+nbits/8;i++){
				biterrs += Bitcnt[data[i]^decdata[i]];
				err += Bitcnt[data[i]^decdata[i]];
			}
			/* here goes list checking */
			
			ii=0;
			while(++ii < L){
			   viterbi(metric,decdata,symbols,nbits,mettab,ii);
			   if(memcmp(data,decdata,1+nbits/8) == 0){
			  printf("\n\n\nFOUND!!!!!\nMetric %d   ",metric[ii]);
			       printf("frame %d path %d decoded data:\n",t,ii);
			       for(i=0;i<1+nbits/8;i++)
				 printf("%02x",decdata[i]);
			       printf("\n");
			       corrected_frames++ ;
			       corrected_bits += err;
			       break;
			   }
			   else {
			  printf("\nNot yet.\nMetric %d   ",metric[ii]);
			       printf("frame %d path %d decoded data:\n",t,ii);
			       for(i=0;i<1+nbits/8;i++)
				 printf("%02x",decdata[i]);
			       printf("\n");
			   }
			}
			
		}
	}
printf("Seed %ld Amplitude %d units, Eb/N0 = %lg dB\n",seed,amp,ebn0);
printf("Amplitude %d units, Eb/N0 = %lg dB  Es/N0 = %lg dB\n",amp,ebn0,esn0);
printf("Frame length = %d bits, #frames = %ld, bits decoded = %ld\n",
       nbits,ntrials,nbits*ntrials);
printf("frame errors: %ld (%g)\n",framerrs,(double)framerrs/ntrials);
printf("Corrected frame errors: %ld (uncorrected %ld)\n",corrected_frames,
       framerrs-corrected_frames);
printf("bit errors: %ld (%g)\n",biterrs,(double)biterrs/(nbits*ntrials));
printf("Corrected bit errors: %ld (uncorrected %ld)\n",corrected_bits,
       biterrs-corrected_bits);
}

usage()
{
	printf("Usage: vittest [options]\n");
	printf("Option&default	meaning\n");
	printf("-a 100		amplitude\n");
	printf("-e 5.0		Eb/N0 in dB\n");
	printf("-n 1152		trial frame size, bits\n");
	printf("-N 10		number of trials\n");
	printf("-r 8		code rate (0->8/9; 8->8/24)\n");
	printf("-s [cur time]	seed for random number generator\n\n");
	printf("-t		select timetest mode (default off)\n");
}
